// CRW - New
#include "WebServer.h"

#include "mongoose.h"

#include "console/console.h"
#include "sim/simConsoleEvent.h"
#include "sim/simBase.h"

#include "Winsock2.h"

int WebServer::requestHandler(struct mg_connection* conn)
{
  char clientIP[60];
  dSprintf(clientIP, 60, "%s:%i", conn->remote_ip, conn->remote_port);

  if (conn->is_websocket) 
  {
    //Con::printf("Websocket request from %s: len=%d bits=%d", clientIP, conn->content_len, conn->wsbits);
    if (conn->content_len )
    {
      char* message = Con::getArgBuffer(conn->content_len+1);
      dStrncpy(message, conn->content, conn->content_len);
      message[conn->content_len] = '\0';
      //Con::printf("Websocket message from %s: %s", clientIP, message);
      if ( Con::isFunction("WebOnWebSocketMessage") )
      {
        const char* result = Con::executef(3, "WebOnWebSocketMessage", clientIP, message);
        if ( dAtob(result) )
          return MG_CLIENT_CONTINUE;
        else
          return MG_CLIENT_CLOSE;
      }
      else
        return MG_CLIENT_CONTINUE;
    }
    else
    {
      return MG_CLIENT_CONTINUE;
    }
  }
  else // Standard request
  {
    if ( Con::isFunction("WebOnNewRequest") )
    {
      // TODO send the request content (for a POST)
      std::string result = Con::executef(5, "WebOnNewRequest", 
        replaceNullWithEmpty(conn->request_method), replaceNullWithEmpty(conn->uri), 
        replaceNullWithEmpty(conn->query_string), clientIP);
      // TODO check the result for some control codes to handle errors
      mg_send_header(conn, "Content-Type", "text/plain");
      mg_send_data(conn, result.c_str(), result.size());
      return MG_REQUEST_PROCESSED;
    }
    else
      return MG_REQUEST_NOT_PROCESSED;
  }
}
int WebServer::connUpdateHandler(mg_connection* conn)
{
  if ( conn->is_websocket )
  {
    WebServer* server = WebServer::webServer();
    // Check to see if we already know about this connection
    auto cIter = server->mWebSocketConnections.find(conn);
    if ( cIter == server->mWebSocketConnections.end() )
    {
      //Con::printf("New connection: %s:%d", conn->remote_ip, conn->remote_port);
      WebServer::connData data;
      data.lastUpdate = Sim::getCurrentTime();
      data.disconnected = false;
      char clientIP[60];
      dSprintf(clientIP, 60, "%s:%i", conn->remote_ip, conn->remote_port);
      data.clientIP = clientIP;
      server->mWebSocketConnections.insert(make_pair(conn, data));
      if ( Con::isFunction("WebOnWebSocketConnect") )
      {
        const char* retVal = Con::executef(5, "WebOnWebSocketConnect", 
          replaceNullWithEmpty(conn->request_method), replaceNullWithEmpty(conn->uri),
          replaceNullWithEmpty(conn->query_string), clientIP);
        if ( !dAtob(retVal) )
          server->closeConnection(string(clientIP));
      }
    }
    else
      cIter->second.lastUpdate = Sim::getCurrentTime();
  }
  return MG_REQUEST_PROCESSED;
}
WebServer* WebServer::theWebServer = nullptr;
WebServer* WebServer::webServer() 
{
  if ( !theWebServer )
    theWebServer = new WebServer();
  return theWebServer;
}

void WebServer::startServer(string& webRoot, string& listenPort)
{
  mMongooseServer = mg_create_server(nullptr);
  mg_set_option(mMongooseServer, "listening_port", listenPort.c_str());
  mg_set_option(mMongooseServer, "document_root", webRoot.c_str());
  mg_set_option(mMongooseServer, "enable_directory_listing", "no");
  mg_set_option(mMongooseServer, "request_timeout_ms", "180000000");

  mg_set_request_handler(mMongooseServer, WebServer::requestHandler);

  setProcessTicks(true);
}
void WebServer::stopServer()
{
  if( mMongooseServer )
    mg_destroy_server(&mMongooseServer);
  setProcessTicks(false);
}
string WebServer::webServerURL()
{
  struct hostent* host = gethostbyname("localhost");
  host = gethostbyname(host->h_name);
  struct in_addr addr;
  addr.s_addr = *(u_long*) host->h_addr_list[0];
  const char* hostname = inet_ntoa(addr);
  const char* port = mg_get_option(mMongooseServer, "listening_port");
  
  string retVal = "http://";
  retVal += hostname;
  retVal += ":";
  retVal += port;
  return retVal;
}
bool WebServer::sendDataToConnection(string &clientIP, const char* data)
{
  // Find the right connection
  for ( auto cIter = mWebSocketConnections.begin(); cIter != mWebSocketConnections.end(); ++cIter )
  {
    if ( cIter->second.clientIP == clientIP && !cIter->second.disconnected )
    {
      if ( cIter->first->is_websocket )
      {
        //Con::printf("Sending websocket message to %s: %s", clientIP.c_str(), data);
        mg_websocket_write(cIter->first, 1, data, dStrlen(data));
        return true;
      }
      else
        return false;
    }
  }
  return false;
}
void WebServer::closeConnection(string &clientIP)
{
  for ( auto cIter = mWebSocketConnections.begin(); cIter != mWebSocketConnections.end(); ++cIter )
  {
    if ( cIter->second.clientIP == clientIP )
    {
      mg_websocket_write(cIter->first, 8, nullptr, 0);
      cIter->second.disconnected = true;
    }
  }
}
void WebServer::interpolateTick( F32 delta )
{
  mg_poll_server(mMongooseServer, 0);
  mg_iterate_over_connections(mMongooseServer, WebServer::connUpdateHandler, nullptr);

  for ( auto cIter = mWebSocketConnections.begin(); cIter != mWebSocketConnections.end(); )
  {
    if ( (unsigned)cIter->second.lastUpdate + 1000 < Sim::getCurrentTime() )
    {
      string clientIP = cIter->second.clientIP;
      bool disconnected = cIter->second.disconnected;
      mWebSocketConnections.erase(cIter++);
      //Con::printf("Connection closed from: %s", clientIP.c_str());
      if ( !disconnected && Con::isFunction("WebOnWebSocketClose") )
        Con::executef(2, "WebOnWebSocketClose", clientIP.c_str());
    }
    else
      ++cIter;
  }
}
#include "web/WebServer_ScriptBinding.h"
